﻿using System;
using System.Drawing;
using Styx.WoWInternals.WoWObjects;
using System.Linq;
using IronBreaker.Core;
using IronBreaker.Core.Helpers;
using IronBreaker.Core.Managers;
using IronBreaker.Core.Utilities;
using Styx;
using Styx.CommonBot;
using Styx.WoWInternals;
using Styx.WoWInternals.DBC;

namespace IronBreaker
{
    public class WoWContextEventArg : EventArgs
    {
        public readonly WoWContext CurrentContext;
        public readonly WoWContext PrevContext;

        public WoWContextEventArg(WoWContext current, WoWContext previous)
        {
            CurrentContext = current;
            PrevContext = previous;
        }
    }

    partial class Main
    {
        #region Declarations

        public static event EventHandler<WoWContextEventArg> ContextChanged;
        private bool _contextEventSubscribed;

        internal static WoWContext ImplicitContext { get; set; }
        private static WoWContext lastContext = WoWContext.None;
        internal static WoWContext _cachedContext = WoWContext.None;
        public static WoWContext TrainingDummyBehaviors { get; set; }
        private static uint lastMap;

        #region Bot Base Contexts
        internal static bool IsQuestBot { get; set; }
        internal static bool IsBattlegroundsBot { get; set; }
        internal static bool IsDungeonBot { get; set; }
        internal static bool IsPokeBot { get; set; }
        internal static bool IsCombatBot { get; set; }
        internal static bool IsGrindBot { get; set; }
        internal static bool IsGatherBot { get; set; }
        internal static bool IsFishingBot { get; set; }
        #endregion

        #endregion

        internal static WoWContext CurrentContext
        {
            get { return _cachedContext; }
            set { _cachedContext = value; }
        }

        public static void GetCurrentContext()
        {
            using (StyxWoW.Memory.AcquireFrame())
            {
                CurrentContext = _GetCurrentContext();
            }
        }

        public static WoWContext _GetCurrentContext()
        {
            if (!StyxWoW.IsInGame)
                return WoWContext.None;
            if (ImplicitContext != WoWContext.None)
            {
                if (lastContext != ImplicitContext)
                    Log.Write(Coloring.HighNormal, "Current Context: {0}", ImplicitContext);
                return ImplicitContext;
            }

            Map map = StyxWoW.Me.CurrentMap;

            if (map.IsBattleground || map.IsArena)
            {
                if (lastContext != WoWContext.Battlegrounds)
                    Log.Write(Coloring.HighNormal, "Current Context: {0}", WoWContext.Battlegrounds);
                return WoWContext.Battlegrounds;
            }

            if (Me.IsInGroup())
            {
                if (Me.IsInInstance)
                {
                    if (lastContext != WoWContext.Instances)
                        Log.Write(Coloring.HighNormal, "Current Context: {0}", WoWContext.Instances);
                    return WoWContext.Instances;
                }

                const WoWPartyMember.GroupRole hasGroupRoleMask = WoWPartyMember.GroupRole.Tank | WoWPartyMember.GroupRole.Damage;

                if ((Me.Role & hasGroupRoleMask) != WoWPartyMember.GroupRole.None)
                {
                    if (lastContext != WoWContext.Instances)
                    {
                        Log.Write(Coloring.HighNormal, "Current Context: {0}", WoWContext.Instances);
                        Log.Write(Coloring.HighNormal, "Current Role: {0}", Me.Role & hasGroupRoleMask);
                    }
                    return WoWContext.Instances;
                }


                if (lastContext != WoWContext.Normal)
                    Log.Write(Coloring.HighNormal, "Current Context: Normal");
                return WoWContext.Normal;

            }

            if (lastContext != WoWContext.Normal)
                Log.Write(Coloring.HighNormal, "Current Context: Normal");
            return WoWContext.Normal;
        }

        private void UpdateContexts()
        {
            if (!_contextEventSubscribed)
            {
                BotEvents.Battleground.OnBattlegroundEntered += e => UpdateContexts();

                OnBotEvent += (src, arg) =>
                {
                    if (arg.Event == IronBreakerBotEvent.BotStarted || arg.Event == IronBreakerBotEvent.BotChanged)
                    {
                        if (UpdateContextStateValues())
                            RebuildBehaviors();
                    }
                };

                _contextEventSubscribed = true;
            }

            GetCurrentContext();

            if (CurrentContext == WoWContext.None)
                return;

            if (CurrentContext != lastContext && ContextChanged != null)
            {
                UpdateContextStateValues();
                WriteContext();

                try
                {
                    ContextChanged(this, new WoWContextEventArg(CurrentContext, lastContext));
                }
                catch
                {
                    // ignored
                }

                lastContext = CurrentContext;
                lastMap = Me.MapId;
            }
            else if (lastMap != Me.MapId)
            {
                WriteContext();
                lastMap = Me.MapId;
            }
        }

        private static bool UpdateContextStateValues()
        {
            bool questBot = IsBotInUse("Quest");
            bool bgBot = IsBotInUse("BGBuddy,", "BG Bot");
            bool dungeonBot = IsBotInUse("DungeonBuddy");
            bool pokeBot = IsBotInUse("Pokébuddy", "PokehBuddy");
            bool manualBot = IsBotInUse("LazyRaider", "Raid Bot", "Tyrael");

            BotBase bot = GetCurrentBotBase();
            bool grindBot = bot != null && bot.Name.ToUpper().Contains("GRIND");
            bool changed = false;

            if (questBot != IsQuestBot)
            {
                changed = true;
                IsQuestBot = questBot;
            }

            if (bgBot != IsBattlegroundsBot)
            {
                changed = true;
                IsBattlegroundsBot = bgBot;
            }

            if (dungeonBot != IsDungeonBot)
            {
                changed = true;
                IsDungeonBot = dungeonBot;
            }

            if (pokeBot != IsPokeBot)
            {
                changed = true;
                IsPokeBot = pokeBot;
            }

            if (manualBot != IsCombatBot)
            {
                changed = true;
                IsCombatBot = manualBot;
            }

            if (grindBot != IsGrindBot)
            {
                changed = true;
                IsGrindBot = grindBot;
            }

            return changed;
        }

        public static void WriteContext()
        {
            using (StyxWoW.Memory.AcquireFrame())
            {
                string sRace = GetRace();
                if (Me.Race == WoWRace.Pandaren)
                    sRace = " " + Me.FactionGroup.ToString() + sRace;

                Log.Write(" ");
                Log.Write(Color.LightSkyBlue, string.Format($"Current Level: {Me.Level}"));
                Log.Write(Color.LightSkyBlue, string.Format($"Current Race: {GetRace()}"));
                Log.Write(Color.LightSkyBlue, string.Format($"Current Spec: {Me.Specialization}"));
                if (IS.EnableDebugLog)
                {
                    Log.Write(Color.LightSkyBlue,
                        string.Format($"Current Bot Running: {GetBotName()} in Zone: {Me.RealZoneText}"));
                    Log.Write(string.Format($"Map Number: {Me.MapId}"));
                    Log.Write(string.Format($"Zone #: {Me.ZoneId}"));
                }

                string runningAs;

                if (Me.CurrentMap == null) { runningAs = "Unknown"; }
                else if (Me.CurrentMap.IsArena) { runningAs = "Arena"; }
                else if (Me.CurrentMap.IsBattleground) { runningAs = "Battleground"; }
                else if (Me.CurrentMap.IsScenario) { runningAs = "Scenario"; }
                else if (Me.CurrentMap.IsRaid) { runningAs = "Raid"; }
                else if (Me.CurrentMap.IsDungeon) { runningAs = "Dungeon"; }
                else if (Me.CurrentMap.IsInstance) { runningAs = "Instance"; }
                else runningAs = "Zone: " + Me.CurrentMap.Name;

                Log.Write(Color.LightSkyBlue, "Running As: {0}", runningAs);
                Log.Write(Color.LightSkyBlue, "Current Context: {0}", CurrentContext == WoWContext.Normal ? "Solo" : CurrentContext.ToString().ToUpper());
                Log.Write(Color.LightSkyBlue, "Running Behaviors: {0}", !Me.IsInGroup() ? "Solo Behaviors" : "Group Behaviors " + Unit.GroupMembers.Count());

                //Log.Write(Color.LightGreen, "Running as: {0} In Context Of: {1} With Behaviors: {2}",
                //    runningAs,
                //    CurrentContext == WoWContext.Normal ? "Solo" : CurrentContext.ToString().ToUpper(),
                //    !Me.IsInGroup() ? "Alone" : "Grouped " + Unit.GroupMembers.Count());

                if (CurrentContext != WoWContext.Battlegrounds && Me.IsInGroup())
                {
                    Log.Write(Color.LightSkyBlue, "INSTANCE GROUP");
                    Log.Write(Color.LightSkyBlue, "ROLE: {0}", (Me.Role & (WoWPartyMember.GroupRole.Tank | WoWPartyMember.GroupRole.Damage)).ToString().ToUpper());
                    Log.Write(Color.LightSkyBlue, "Current Party Size: {0} / {1}", Me.GroupInfo.NumRaidMembers, (int)Math.Max(Me.CurrentMap.MaxPlayers, Me.GroupInfo.GroupSize));

                    //Log.Write(Color.LightGreen, "We are grouped up with the role: {0} With: {1} of {2} players",
                    //    (Me.Role & (WoWPartyMember.GroupRole.Tank | WoWPartyMember.GroupRole.Damage)).ToString().ToUpper(),
                    //    Me.GroupInfo.NumRaidMembers,
                    //    (int) Math.Max(Me.CurrentMap.MaxPlayers, Me.GroupInfo.GroupSize));
                }
            }
        }

        public static bool InMovie()
        {
            bool cutscene = Lua.GetReturnVal<bool>("return InCinematic()", 0);
            return cutscene;
        }

        #region [Methods] - Get Current Bot Properties

        public static string GetBotName()
        {
            BotBase bot = null;

            if (TreeRoot.Current != null)
            {
                if (!(TreeRoot.Current is NewMixedMode.MixedModeEx))
                    bot = TreeRoot.Current;
                else
                {
                    NewMixedMode.MixedModeEx mmb = (NewMixedMode.MixedModeEx)TreeRoot.Current;
                    if (mmb != null)
                    {
                        if (mmb.SecondaryBot != null && mmb.SecondaryBot.RequirementsMet)
                            return "Mixed: " + mmb.SecondaryBot.Name;
                        return mmb.PrimaryBot != null ? "Mixed: " + mmb.PrimaryBot.Name : "Mixed: Primary[NULL]";
                    }
                }
            }
            return bot == null ? "(null)" : bot.Name;
        }

        public static BotBase GetCurrentBotBase()
        {
            BotBase bot = TreeRoot.Current;

            if (bot is NewMixedMode.MixedModeEx)
            {
                NewMixedMode.MixedModeEx mmb = (NewMixedMode.MixedModeEx)bot;

                if (mmb != null)
                {
                    if (mmb.SecondaryBot != null && mmb.SecondaryBot.RequirementsMet)
                        bot = mmb.SecondaryBot;
                    else
                        bot = mmb.PrimaryBot;
                }
            }
            return bot;
        }

        public static bool IsBotInUse(params string[] nameSubStrings)
        {
            string botName = GetBotName().ToUpper();
            return nameSubStrings.Any(s => botName.Contains(s.ToUpper()));
        }

        #endregion

        #region [Methods] - Get Character Name Properties

        public static string GetRace()
        {
            return Me.Race.ToString().CamelFormat();
        }

        public static string GetClass()
        {
            return Me.Class.ToString().CamelFormat();
        }

        public static string GetSpecialization()
        {
            if (TalentManagement.CurrentSpec == WoWSpec.None)
                return "No Spec Found";

            string classSpec = TalentManagement.CurrentSpec.ToString();
            classSpec = classSpec.Substring(Me.Class.ToString().Length);
            return " " + classSpec;
        }

        public static string GetSpecAndClass()
        {
            return GetSpecialization() + Me.Class.ToString().CamelFormat();
        }

        public static string GetClassAndSpec()
        {
            return Me.Class.ToString().CamelFormat() + GetSpecialization();
        }

        #endregion
    }
}
